import logging

from pydantic import BaseModel
from sqlalchemy import func
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select

from service_repository.filters.filters import apply_filters
from service_repository.filters.pagination import apply_pagination
from service_repository.filters.sorting import apply_sort
from service_repository.interfaces.repository import RepositoryInterface

logger = logging.getLogger(__name__)


class BaseRepositorySqlalchemy(RepositoryInterface):
    """Class representing the SQLAlchemy abstract repository."""

    _model = None
    max_per_page = 25

    def __init__(self, db: AsyncSession) -> None:
        self.db: AsyncSession = db

    async def create(self, schema_in: dict, autocommit: bool = True):
        """Create new object and returns the saved object instance."""
        instance = self.model(**schema_in)
        self.db.add(instance)
        if autocommit:
            await self.db.commit()
            await self.db.refresh(instance)
        return instance

    async def update(
        self, instance: BaseModel, schema_in: dict, autocommit: bool = True
    ):
        """Update a instance."""
        data = instance.dict()

        for field in data:
            if field in schema_in:
                setattr(instance, field, schema_in[field])

        if autocommit:
            await self.db.commit()
        return instance

    async def get(self, **kwargs):
        """Get one instance by filter."""
        query = await self.db.execute(select(self.model).filter_by(**kwargs))
        instance = query.scalar_one_or_none()
        await self.db.commit()

        if instance:
            return instance

    async def all(self, **kwargs):
        """Get all instances by filter."""
        query = await self.db.execute(select(self.model).filter_by(**kwargs))
        instances = query.scalars().all()
        await self.db.commit()
        return instances

    async def paginate(
        self,
        page: int = 1,
        per_page: int = 15,
        criteria: dict = {},
        sort: list = [],
    ):
        """Get collection of instances paginated by filter."""
        if per_page == -1:
            per_page = None
        elif per_page > self.max_per_page:
            per_page = self.max_per_page

        stmt = select(self.model)

        if criteria:
            stmt = apply_filters(stmt, criteria)

        if sort:
            stmt = apply_sort(stmt, sort)

        stmt, pagination = await apply_pagination(
            stmt,
            session=self.db,
            model=self.model,
            page_number=page,
            page_size=per_page,
        )

        query = await self.db.execute(stmt)
        await self.db.commit()

        response = {
            "items": query.scalars().all(),
            "per_page": per_page,
            "num_pages": pagination.num_pages,
            "page": pagination.page_number,
            "total": pagination.total_results,
        }

        return response

    async def delete(self, **kwargs):
        """Delete one instance by filter."""
        instance = await self.get(**kwargs)
        await self.db.delete(instance)
        await self.db.commit()

    async def count(self, **kwargs):
        """Count instances by filter."""
        primary_key = getattr(
            self.model,
            self.model.__table__.primary_key.columns_autoinc_first[0].name,
        )

        stmt = select(self.model)

        count = await self.db.execute(
            stmt.with_only_columns(func.count(primary_key)).filter_by(**kwargs)
        )

        total = count.scalar_one()
        return total

    @property
    def model(self):
        if self._model is None:
            raise ValueError("Model is None, set the model")
        return self._model

    @model.setter
    def model(self, value):
        self._model = value



import math

from motor.motor_asyncio import AsyncIOMotorDatabase
from pydantic import BaseModel

from service_repository.interfaces.repository import RepositoryInterface


class BaseRepositoryMotor(RepositoryInterface):
    """Class representing the motor abstract repository."""

    _model = None
    _collection = None

    def __init__(self, db: AsyncIOMotorDatabase) -> None:
        self.db: AsyncIOMotorDatabase = db

    async def create(self, schema_in: dict):
        """Create new object and returns the saved object instance."""
        create_data = self.model(**schema_in).dict()
        collection = self.db.get_collection(self.collection)
        result = await collection.insert_one(create_data)
        instance = await collection.find_one({"_id": result.inserted_id})
        schema_out = self.model(**instance)
        return schema_out

    async def update(self, instance: BaseModel, schema_in: dict):
        """Update a instance."""
        update_data = {
            "$set": schema_in,
            "$currentDate": {"updated_at": True},
        }
        criteria = {"_id": instance.id}
        collection = self.db.get_collection(self.collection)
        await collection.update_one(criteria, update_data)
        instance = await collection.find_one(criteria)
        schema_out = self.model(**instance)
        return schema_out

    async def get(self, **kwargs):
        """Get one instance by filter."""
        collection = self.db.get_collection(self.collection)
        instance = await collection.find_one(kwargs)
        if instance:
            schema_out = self.model(**instance)
            return schema_out

    async def delete(self, **kwargs):
        """Delete one instance by filter."""
        collection = self.db.get_collection(self.collection)
        await collection.delete_one(kwargs)

    async def count(self, **kwargs):
        """Count instances by filter."""
        collection = self.db.get_collection(self.collection)
        total = await collection.count_documents(kwargs)
        return total

    async def all(self, **kwargs):
        """Count instances by filter."""
        collection = self.db.get_collection(self.collection)
        total = await collection.count_documents(kwargs)
        items = collection.find(kwargs)
        instances = await items.to_list(total)
        return instances

    async def paginate(
        self,
        page: int = 1,
        per_page: int = 15,
        criteria: dict = {},
        sort: list = [],
    ):
        """Get collection of instances paginated by filter."""
        collection = self.db.get_collection(self.collection)
        total = await collection.count_documents(criteria)
        items = collection.find(criteria)

        if sort:
            items = items.sort(sort)

        items = await items.to_list(per_page)

        response = {
            "items": [self.model(**item) for item in items],
            "per_page": per_page,
            "num_pages": int(math.ceil(total / per_page)),
            "page": page,
            "total": total,
        }
        return response

    @property
    def model(self):
        if self._model is None:
            raise ValueError("Model is None, set the model")
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def collection(self):
        if self._collection is None:
            raise ValueError("Collection is None, set the collection")
        return self._collection

    @collection.setter
    def collection(self, value):
        self._collection = value



class BadFilterFormat(Exception):
    pass


class BadSortFormat(Exception):
    pass


class BadLoadFormat(Exception):
    pass


class BadSpec(Exception):
    pass


class FieldNotFound(Exception):
    pass


class BadQuery(Exception):
    pass


class InvalidPage(Exception):
    pass



import types

from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.util import symbol

from .exceptions import BadQuery, BadSpec, FieldNotFound


class Field(object):
    def __init__(self, model, field_name):
        self.model = model
        self.field_name = field_name

    def get_sqlalchemy_field(self):
        if self.field_name not in self._get_valid_field_names():
            raise FieldNotFound(
                "Model {} has no column `{}`.".format(
                    self.model, self.field_name
                )
            )
        sqlalchemy_field = getattr(self.model, self.field_name)

        # If it's a hybrid method, then we call it so that we can work with
        # the result of the execution and not with the method object itself
        if isinstance(sqlalchemy_field, types.MethodType):
            sqlalchemy_field = sqlalchemy_field()

        return sqlalchemy_field

    def _get_valid_field_names(self):
        inspect_mapper = inspect(self.model)
        columns = inspect_mapper.columns
        orm_descriptors = inspect_mapper.all_orm_descriptors

        column_names = columns.keys()
        hybrid_names = [
            key
            for key, item in orm_descriptors.items()
            if _is_hybrid_property(item) or _is_hybrid_method(item)
        ]

        return set(column_names) | set(hybrid_names)


def _is_hybrid_property(orm_descriptor):
    return orm_descriptor.extension_type == symbol("HYBRID_PROPERTY")


def _is_hybrid_method(orm_descriptor):
    return orm_descriptor.extension_type == symbol("HYBRID_METHOD")


def get_query_models(query):
    """Get models from query.
    :param query:
        A :class:`sqlalchemy.orm.Query` instance.
    :returns:
        A dictionary with all the models included in the query.
    """
    models = [col_desc["entity"] for col_desc in query.column_descriptions]
    # if hasattr(query, "_join_entities"):  # sqlalchemy<1.4
    #     models.extend(mapper.class_ for mapper in query._join_entities)
    # else:  # sqlalchemy>=1.4
    #     models.extend(
    #         mapper.class_ for mapper in query._compile_state()._join_entities
    #     )

    # account also query.select_from entities
    if hasattr(query, "_select_from_entity") and (
        query._select_from_entity is not None
    ):
        model_class = (
            query._select_from_entity.class_
            if isinstance(query._select_from_entity, Mapper)  # sqlalchemy>=1.1
            else query._select_from_entity  # sqlalchemy==1.0
        )
        if model_class not in models:
            models.append(model_class)

    return {model.__name__: model for model in models}


def get_model_from_spec(spec, query, default_model=None):
    """Determine the model to which a spec applies on a given query.
    A spec that does not specify a model may be applied to a query that
    contains a single model. Otherwise the spec must specify the model to
    which it applies, and that model must be present in the query.
    :param query:
        A :class:`sqlalchemy.orm.Query` instance.
    :param spec:
        A dictionary that may or may not contain a model name to resolve
        against the query.
    :returns:
        A model instance.
    :raise BadSpec:
        If the spec is ambiguous or refers to a model not in the query.
    :raise BadQuery:
        If the query contains no models.
    """
    models = get_query_models(query)
    if not models:
        raise BadQuery("The query does not contain any models.")

    model_name = spec.get("model")
    if model_name is not None:
        models = [v for (k, v) in models.items() if k == model_name]
        if not models:
            raise BadSpec(
                "The query does not contain model `{}`.".format(model_name)
            )
        model = models[0]
    else:
        if len(models) == 1:
            model = list(models.values())[0]
        elif default_model is not None:
            return default_model
        else:
            raise BadSpec("Ambiguous spec. Please specify a model.")

    return model


def get_model_class_by_name(registry, name):
    """Return the model class matching `name` in the given `registry`."""
    for cls in registry.values():
        if getattr(cls, "__name__", None) == name:
            return cls


def get_default_model(query):
    """Return the singular model from `query`, or `None` if `query` contains
    multiple models.
    """
    return None
    query_models = get_query_models(query).values()
    if len(query_models) == 1:
        (default_model,) = iter(query_models)
    else:
        default_model = None
    return default_model


def auto_join(query, *model_names):
    """Automatically join models to `query` if they're not already present
    and the join can be done implicitly.
    """
    # every model has access to the registry, so we can use any from the query
    return query
    query_models = get_query_models(query).values()
    model = list(query_models)[-1]
    if hasattr(model, "_decl_class_registry"):  # sqlalchemy<1.4
        model_registry = model._decl_class_registry
    else:  # sqlalchemy>=1.4
        model_registry = model.registry._class_registry

    for name in model_names:
        model = get_model_class_by_name(model_registry, name)
        if model not in get_query_models(query).values():
            try:
                query = query.join(model)
            except InvalidRequestError:
                pass  # can't be autojoined
    return query



import math
from collections import namedtuple

from sqlalchemy import func
from sqlalchemy.ext.asyncio import AsyncSession
from sqlmodel.main import SQLModel

from service_repository.exceptions import InvalidPage


async def apply_pagination(
    stmt,
    session: AsyncSession,
    model: SQLModel,
    page_number=None,
    page_size=None,
):
    """Apply pagination to a SQLAlchemy stmt object.

    :param stmt:
        A :class:`sqlalchemy.sql.selectable.Select` instance.

    :param session:
        A :class:`sqlalchemy.ext.asyncio.AsyncSession` instance.

    :param model:
        A :class:`sqlmodel.main.SQLModel` class.

    :param page_number:
        Page to be returned (starts and defaults to 1).

    :param page_size:
        Maximum number of results to be returned in the page (defaults
        to the total results).

    :returns:
        A 2-tuple with the paginated SQLAlchemy stmt object and
        a pagination namedtuple.

        The pagination object contains information about the results
        and pages: ``page_size`` (defaults to ``total_results``),
        ``page_number`` (defaults to 1), ``num_pages`` and
        ``total_results``.

    Basic usage::

        stmt, pagination = apply_pagination(stmt, 1, 10)
        >>> len(stmt)
        10
        >>> pagination.page_size
        10
        >>> pagination.page_number
        1
        >>> pagination.num_pages
        3
        >>> pagination.total_results
        22
        >>> page_size, page_number, num_pages, total_results = pagination
    """
    primary_key = getattr(
        model, model.__table__.primary_key.columns_autoinc_first[0].name
    )

    query_count = await session.execute(
        stmt.with_only_columns(func.count(primary_key))
    )

    total_results = query_count.scalar_one()
    stmt = _limit(stmt, page_size)

    # Page size defaults to total results
    if page_size is None:
        page_size = total_results

    stmt = _offset(stmt, page_number, page_size)

    # Page number defaults to 1
    if page_number is None:
        page_number = 1

    num_pages = _calculate_num_pages(page_size, total_results)

    Pagination = namedtuple(
        "Pagination",
        ["page_number", "page_size", "num_pages", "total_results"],
    )
    return stmt, Pagination(page_number, page_size, num_pages, total_results)


def _limit(query, page_size):
    if page_size is not None:
        if page_size < 0:
            raise InvalidPage(
                "Page size should not be negative: {}".format(page_size)
            )

        query = query.limit(page_size)

    return query


def _offset(query, page_number, page_size):
    if page_number is not None:
        if page_number < 1:
            raise InvalidPage(
                "Page number should be positive: {}".format(page_number)
            )

        query = query.offset((page_number - 1) * page_size)

    return query


def _calculate_num_pages(page_size, total_results):
    if page_size == 0:
        return 0

    return math.ceil(float(total_results) / float(page_size))



# -*- coding: utf-8 -*-

from .exceptions import BadSortFormat
from .models import Field, auto_join, get_default_model, get_model_from_spec

SORT_ASCENDING = "asc"
SORT_DESCENDING = "desc"


class Sort(object):
    def __init__(self, sort_spec):
        self.sort_spec = sort_spec

        try:
            field_name = sort_spec["field"]
            direction = sort_spec["direction"]
        except KeyError:
            raise BadSortFormat(
                "`field` and `direction` are mandatory attributes."
            )
        except TypeError:
            raise BadSortFormat(
                "Sort spec `{}` should be a dictionary.".format(sort_spec)
            )

        if direction not in [SORT_ASCENDING, SORT_DESCENDING]:
            raise BadSortFormat("Direction `{}` not valid.".format(direction))

        self.field_name = field_name
        self.direction = direction
        self.nullsfirst = sort_spec.get("nullsfirst")
        self.nullslast = sort_spec.get("nullslast")

    def get_named_models(self):
        if "model" in self.sort_spec:
            return {self.sort_spec["model"]}
        return set()

    def format_for_sqlalchemy(self, query, default_model):
        sort_spec = self.sort_spec
        direction = self.direction
        field_name = self.field_name

        model = get_model_from_spec(sort_spec, query, default_model)

        field = Field(model, field_name)
        sqlalchemy_field = field.get_sqlalchemy_field()

        if direction == SORT_ASCENDING:
            sort_fnc = sqlalchemy_field.asc
        elif direction == SORT_DESCENDING:
            sort_fnc = sqlalchemy_field.desc

        if self.nullsfirst:
            return sort_fnc().nullsfirst()
        elif self.nullslast:
            return sort_fnc().nullslast()
        else:
            return sort_fnc()


def get_named_models(sorts):
    models = set()
    for sort in sorts:
        models.update(sort.get_named_models())
    return models


def apply_sort(query, sort_spec):
    """Apply sorting to a :class:`sqlalchemy.orm.Query` instance.

    :param sort_spec:
        A list of dictionaries, where each one of them includes
        the necesary information to order the elements of the query.

        Example::

            sort_spec = [
                {'model': 'Foo', 'field': 'name', 'direction': 'asc'},
                {'model': 'Bar', 'field': 'id', 'direction': 'desc'},
                {
                    'model': 'Qux',
                    'field': 'surname',
                    'direction': 'desc',
                    'nullslast': True,
                },
                {
                    'model': 'Baz',
                    'field': 'count',
                    'direction': 'asc',
                    'nullsfirst': True,
                },
            ]

        If the query being modified refers to a single model, the `model` key
        may be omitted from the sort spec.

    :returns:
        The :class:`sqlalchemy.orm.Query` instance after the provided
        sorting has been applied.
    """
    if isinstance(sort_spec, dict):
        sort_spec = [sort_spec]

    sorts = [Sort(item) for item in sort_spec]

    default_model = get_default_model(query)

    sort_models = get_named_models(sorts)
    query = auto_join(query, *sort_models)

    sqlalchemy_sorts = [
        sort.format_for_sqlalchemy(query, default_model) for sort in sorts
    ]

    if sqlalchemy_sorts:
        query = query.order_by(*sqlalchemy_sorts)

    return query



from abc import ABCMeta, abstractmethod

from pydantic import BaseModel


class ServiceInterface(metaclass=ABCMeta):
    """Class representing the service interface."""

    @abstractmethod
    async def create(self, schema_in: BaseModel):
        """
        Create new entity and returns the saved instance.
        """
        raise NotImplementedError()

    @abstractmethod
    async def update(self, instance: BaseModel, schema_in: BaseModel):
        """Updates an entity and returns the saved instance."""
        raise NotImplementedError()

    @abstractmethod
    async def get(self, **kwargs):
        """Get and return one instance by filter."""
        raise NotImplementedError()

    @abstractmethod
    async def delete(self, **kwargs):
        """Delete one instance by filter."""
        raise NotImplementedError()



from collections import namedtuple
from collections.abc import Iterable
from inspect import signature
from itertools import chain

from six import string_types
from sqlalchemy import and_, func, not_, or_
from sqlalchemy.sql.selectable import Select

from .exceptions import BadFilterFormat
from .models import Field, get_default_model, get_model_from_spec

BooleanFunction = namedtuple(
    "BooleanFunction", ("key", "sqlalchemy_fn", "only_one_arg")
)
BOOLEAN_FUNCTIONS = [
    BooleanFunction("or", or_, False),
    BooleanFunction("and", and_, False),
    BooleanFunction("not", not_, True),
]
"""
Sqlalchemy boolean functions that can be parsed from the filter definition.
"""


class Operator(object):
    OPERATORS = {
        "is_null": lambda f: f.is_(None),
        "is_not_null": lambda f: f.isnot(None),
        "==": lambda f, a: f == a,
        "eq": lambda f, a: f == a,
        "!=": lambda f, a: f != a,
        "ne": lambda f, a: f != a,
        ">": lambda f, a: f > a,
        "gt": lambda f, a: f > a,
        "<": lambda f, a: f < a,
        "lt": lambda f, a: f < a,
        ">=": lambda f, a: f >= a,
        "ge": lambda f, a: f >= a,
        "<=": lambda f, a: f <= a,
        "le": lambda f, a: f <= a,
        "like": lambda f, a: f.like(a),
        "ilike": lambda f, a: f.ilike(a),
        "not_ilike": lambda f, a: ~f.ilike(a),
        "in": lambda f, a: f.in_(a),
        "not_in": lambda f, a: ~f.in_(a),
        "any": lambda f, a: f.any(a),
        "not_any": lambda f, a: func.not_(f.any(a)),
    }

    def __init__(self, operator=None):
        if not operator:
            operator = "=="

        if operator not in self.OPERATORS:
            raise BadFilterFormat("Operator `{}` not valid.".format(operator))

        self.operator = operator
        self.function = self.OPERATORS[operator]
        self.arity = len(signature(self.function).parameters)


class Filter(object):
    def __init__(self, filter_spec):
        self.filter_spec = filter_spec

        try:
            filter_spec["field"]
        except KeyError:
            raise BadFilterFormat("`field` is a mandatory filter attribute.")
        except TypeError:
            raise BadFilterFormat(
                "Filter spec `{}` should be a dictionary.".format(filter_spec)
            )

        self.operator = Operator(filter_spec.get("op"))
        self.value = filter_spec.get("value")
        value_present = True if "value" in filter_spec else False
        if not value_present and self.operator.arity == 2:
            raise BadFilterFormat("`value` must be provided.")

    def get_named_models(self):
        if "model" in self.filter_spec:
            return {self.filter_spec["model"]}
        return set()

    def format_for_sqlalchemy(self, query, default_model):
        filter_spec = self.filter_spec
        operator = self.operator
        value = self.value

        model = get_model_from_spec(filter_spec, query, default_model)

        function = operator.function
        arity = operator.arity

        field_name = self.filter_spec["field"]
        field = Field(model, field_name)
        sqlalchemy_field = field.get_sqlalchemy_field()

        if arity == 1:
            return function(sqlalchemy_field)

        if arity == 2:
            return function(sqlalchemy_field, value)


class BooleanFilter(object):
    def __init__(self, function, *filters):
        self.function = function
        self.filters = filters

    def get_named_models(self):
        models = set()
        for filter in self.filters:
            models.update(filter.get_named_models())
        return models

    def format_for_sqlalchemy(self, query, default_model):
        return self.function(
            *[
                filter.format_for_sqlalchemy(query, default_model)
                for filter in self.filters
            ]
        )


def _is_iterable_filter(filter_spec):
    """`filter_spec` may be a list of nested filter specs, or a dict."""
    return isinstance(filter_spec, Iterable) and not isinstance(
        filter_spec, (string_types, dict)
    )


def build_filters(filter_spec):
    """Recursively process `filter_spec`"""

    if _is_iterable_filter(filter_spec):
        return list(
            chain.from_iterable(build_filters(item) for item in filter_spec)
        )

    if isinstance(filter_spec, dict):
        # Check if filter spec defines a boolean function.
        for boolean_function in BOOLEAN_FUNCTIONS:
            if boolean_function.key in filter_spec:
                # The filter spec is for a boolean-function
                # Get the function argument definitions and validate
                fn_args = filter_spec[boolean_function.key]

                if not _is_iterable_filter(fn_args):
                    raise BadFilterFormat(
                        "`{}` value must be an iterable across the function "
                        "arguments".format(boolean_function.key)
                    )
                if boolean_function.only_one_arg and len(fn_args) != 1:
                    raise BadFilterFormat(
                        "`{}` must have one argument".format(
                            boolean_function.key
                        )
                    )
                if not boolean_function.only_one_arg and len(fn_args) < 1:
                    raise BadFilterFormat(
                        "`{}` must have one or more arguments".format(
                            boolean_function.key
                        )
                    )
                return [
                    BooleanFilter(
                        boolean_function.sqlalchemy_fn, *build_filters(fn_args)
                    )
                ]

    return [Filter(filter_spec)]


def get_named_models(filters):
    models = set()
    for filter in filters:
        models.update(filter.get_named_models())
    return models


def apply_filters(stmt: Select, filter_spec, do_auto_join=True):
    """Apply filters to a SQLAlchemy statement.

    :param stmt:
        A :class:`sqlalchemy.sql.selectable.Select` instance.

    :param filter_spec:
        A dict or an iterable of dicts, where each one includes
        the necesary information to create a filter to be applied to the
        query.

        Example::

            filter_spec = [
                {'model': 'Foo', 'field': 'name', 'op': '==', 'value': 'foo'},
            ]

        If the query being modified refers to a single model, the `model` key
        may be omitted from the filter spec.

        Filters may be combined using boolean functions.

        Example:

            filter_spec = {
                'or': [
                    {'model': 'Foo', 'field': 'id', 'op': '==', 'value': '1'},
                    {'model': 'Bar', 'field': 'id', 'op': '==', 'value': '2'},
                ]
            }

    :returns:
        The :class:`sqlalchemy.sql.selectable.Select` instance after all the
        filters have been applied.
    """
    filters = build_filters(filter_spec)

    default_model = get_default_model(stmt)

    sqlalchemy_filters = [
        filter.format_for_sqlalchemy(stmt, default_model) for filter in filters
    ]

    if sqlalchemy_filters:
        stmt = stmt.where(*sqlalchemy_filters)

    return stmt



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['service_repository',
 'service_repository.filters',
 'service_repository.interfaces',
 'service_repository.repositories']

package_data = \
{'': ['*']}

install_requires = \
['six>=1.16.0,<2.0.0']

setup_kwargs = {
    'name': 'service-repository',
    'version': '0.2.0',
    'description': 'Asynchronous abstract methods for service layer built into the repository pattern for SQLAlchemy or MongoDB with the same service layer.',
    'long_description': '# Abstract service layer built into the repository pattern for SQLAlchemy or MongoDB (asynchronous)\n\nAsynchronous abstract methods for service layer built into the repository pattern for SQLAlchemy or MongoDB with the same service layer.\n\nInherit the usual service methods:\n`pagination`, `create`, `get`, `update`, `delete`, `count`\n\n## Installation\n\n<div class="termy">\n\n```console\n$ pip install service-repository\n\n---> 100%\n```\n\n</div>\n\n## With SQLAlchemy\n\nCreate the model, as in the example [song/models.py](tests/song/models.py)\n\nCreate and extend of `BaseRepositorySqlalchemy` in you repository, as in the example [song/repositories.py](tests/song/repositories.py)\n\nCreate and extend you service, as in the example [song/services.py](tests/song/services.py)\n\nThe following is an example with the connection session, but pass the session as needed.\n\n```python\nimport pytest\nfrom sqlalchemy.ext.asyncio import AsyncSession, create_async_engine\nfrom sqlalchemy.orm import sessionmaker\nfrom sqlmodel import SQLModel\n\n\n# Get you connection session of SQLAlchemy, example next, but make as you\n@pytest.fixture()\nasync def sqlalchemy():\n    SQLALCHEMY_DATABASE_URI = "sqlite+aiosqlite:///./tests/testing.db"\n    engine = create_async_engine(\n        SQLALCHEMY_DATABASE_URI, future=True, echo=True\n    )\n    async_session = sessionmaker(\n        bind=engine, expire_on_commit=False, class_=AsyncSession\n    )\n    async with engine.begin() as conn:\n        await conn.run_sync(SQLModel.metadata.create_all)\n        yield async_session\n        await conn.run_sync(SQLModel.metadata.drop_all)\n```\n\nThe following is an example with register created.\n\n```python\nimport pytest\nfrom tests.song.services import SongService\nfrom tests.song.models import SongCreate\n\n\n@pytest.fixture\nasync def song_one(sqlalchemy):\n    song_create = {\n        "title": "Song title 1",\n        "is_active": True,\n    }\n    \n    async with sqlalchemy() as session:\n        song = await SongService(db=session).create(\n            schema_in=SongCreate(**song_create)\n        )\n    return song\n```\n\n#### Pagination with dynamic filter and sorting in the service.\n\n```python\nimport pytest\nfrom tests.song.services import SongService\n\n\n@pytest.mark.asyncio\nasync def test_song_service_with_or_and_asc_filter_paginate_song(sqlalchemy):\n    criteria = [\n        {"field": "foo", "op": "ilike", "value": "%bar%"},\n    ]\n    # Or with criteria in format or\n    criteria = [\n        {\n            "or": [\n                {"field": "title", "op": "==", "value": "Song title 2"},\n                {"field": "title", "op": "==", "value": "Song title 3"},\n            ]\n        }\n    ]\n    sort = [\n        {"field": "title", "direction": "asc"},\n    ]\n    async with sqlalchemy() as session:\n        pagination = await SongService(db=session).paginate(\n            page=1, per_page=5, criteria=criteria, sort=sort\n        )\n```\n\nOperator options for filtering are:\n\n`is_null, is_not_null, eq, ne, gt, lt, ge, le, like, ilike, not_ilike, in, not_in, any, not_any`\n\n#### Use the create extended method on service\n\n```python\nimport pytest\nfrom tests.song.services import SongService\nfrom tests.song.models import SongCreate\n\n\n@pytest.mark.asyncio\nasync def test_song_service_create_song(sqlalchemy):\n    song_data_one = {\n        "title": "Song title 1",\n        "is_active": True,\n    }\n    async with sqlalchemy() as session:\n        song = await SongService(db=session).create(\n            schema_in=SongCreate(**song_data_one)\n        )\n```\n\n#### Update method on service\n\n```python\nimport pytest\nfrom tests.song.services import SongService\nfrom tests.song.models import SongUpdate\n\n\n@pytest.mark.asyncio\nasync def test_song_service_update_song(sqlalchemy, song_one):\n    song_data = {\n        "title": "Song title 2",\n        "is_active": False,\n    }\n    async with sqlalchemy() as session:\n        song = await SongService(db=session).update(\n            instance=song_one, schema_in=SongUpdate(**song_data)\n        )\n```\n\n#### Get method on service to get one register\n\n```python\nimport pytest\nfrom tests.song.services import SongService\n\n\n@pytest.mark.asyncio\nasync def test_song_service_get_song(sqlalchemy, song_one):\n    async with sqlalchemy() as session:\n        song = await SongService(db=session).get(id=song_one.id)\n```\n\n#### Delete method on service\n\n```python\nimport pytest\nfrom tests.song.services import SongService\n\n\n@pytest.mark.asyncio\nasync def test_song_service_delete_song(sqlalchemy, song_one):\n    async with sqlalchemy() as session:\n        await SongService(db=session).delete(id=song_one.id)\n```\n\n#### Count method on service\n\n```python\nimport pytest\nfrom tests.song.services import SongService\n\n\n@pytest.mark.asyncio\nasync def test_song_service_count_song(sqlalchemy):\n    async with sqlalchemy() as session:\n        total = await SongService(db=session).count()\n```\n\n## With MongoDB\n\nCreate the model, as in the example [product/models.py](tests/product/models.py)\n\nCreate and extend of `BaseRepositoryMotor` in you repository, as in the example [product/repositories.py](tests/product/repositories.py)\n\nCreate and extend you service, as in the example [product/services.py](tests/product/services.py)\n\nThe following is an example with the connection session, but pass the session as needed.\n\n```python\nimport pytest\nfrom motor.motor_asyncio import AsyncIOMotorClient\n\n\n# Get you connection session of Motor, example next, but make as you\n@pytest.fixture()\nasync def motor():\n    client = AsyncIOMotorClient(\n        "localhost",\n        maxPoolSize=10,\n        minPoolSize=10,\n        tz_aware=True,\n    )\n    await client.drop_database("testing")\n    session = client.get_database("testing")\n    yield session\n```\n\nThe following is an example with register created.\n\n```python\nimport pytest\nfrom tests.product.models import ProductCreate\nfrom tests.product.services import ProductService\n\n\n@pytest.fixture\nasync def product_one(motor, product_data_one):\n    product_data = {\n        "title": "Product title 1",\n        "is_active": True,\n    }\n    product = await ProductService(db=motor).create(\n        schema_in=ProductCreate(**product_data)\n    )\n    return product\n```\n\n#### Pagination with dynamic filter and sorting in the service.\n\n```python\nimport pytest\nfrom tests.product.services import ProductService\n\n\n@pytest.mark.asyncio\nasync def test_product_service_with_or_and_asc_filter_paginate_product(\n    app, motor, product_data_one\n):\n    criteria = {"title": "Product title 1"}\n    sort = [("title", -1)]\n\n    pagination = await ProductService(db=motor).paginate(\n        page=1, per_page=5, criteria=criteria, sort=sort\n    )\n```\n\n#### Use the create extended method on service\n\n```python\nimport pytest\nfrom tests.product.models import ProductCreate\nfrom tests.product.services import ProductService\n\n\n@pytest.mark.asyncio\nasync def test_product_service_create_product(motor):\n    product_data = {\n        "title": "Product title 1",\n        "is_active": True,\n    }\n    product = await ProductService(db=motor).create(\n        schema_in=ProductCreate(**product_data)\n    )\n```\n\n#### Update method on service\n\n```python\nimport pytest\nfrom tests.product.models import ProductUpdate\nfrom tests.product.services import ProductService\n\n\n@pytest.mark.asyncio\nasync def test_product_service_update_product(motor, product_one):\n    product_data = {\n        "title": "Product title 2",\n        "is_active": False,\n    }\n    product = await ProductService(db=motor).update(\n        instance=product_one, schema_in=ProductUpdate(**product_data)\n    )\n```\n\n#### Get method on service to get one register\n\n```python\nimport pytest\nfrom tests.product.services import ProductService\n\n\n@pytest.mark.asyncio\nasync def test_product_service_get_product(motor, product_one):\n    product = await ProductService(db=motor).get(_id=product_one.id)\n```\n\n#### Delete method on service\n\n```python\nimport pytest\nfrom tests.product.services import ProductService\n\n\n@pytest.mark.asyncio\nasync def test_product_service_delete_product(motor, product_one):\n    await ProductService(db=motor).delete(_id=product_one.id)\n```\n\n#### Count method on service\n\n```python\nimport pytest\nfrom tests.product.services import ProductService\n\n\n@pytest.mark.asyncio\nasync def test_product_service_count_product(motor):\n    total = await ProductService(db=motor).count()\n```\n\n### Inherited base service layer with FastApi\n\nYou can create an inherited base service layer and add its methods, \nas in the example in FastApi with convenience methods for 404:\n\n```python\nfrom fastapi import HTTPException, status\nfrom pydantic import BaseModel\nfrom service_repository.services import BaseService as Base\n\n\nclass BaseService(Base):\n    """Class representing the base service."""\n    def __init__(self, db) -> None:\n        self.db = db\n\n    async def get_or_404(self, **kwargs):\n        """Get one instance by filter or 404 if not exists."""\n        instance = await self.get(**kwargs)\n        if not instance:\n            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)\n        return instance\n\n    async def update_or_404(self, schema_in: BaseModel, **kwargs):\n        """Update instance or 404 if not exists."""\n        instance = await self.get_or_404(**kwargs)\n        instance = await self.update(instance=instance, schema_in=schema_in)\n        return instance\n\n    async def delete_or_404(self, **kwargs):\n        """Delete one instance by filter or 404 if not exists."""\n        await self.get_or_404(**kwargs)\n        await self.delete(**kwargs)\n```\n\n### Inherited base service layer with Flask\n\nYou can create an inherited base service layer and add its methods, \nas in the example in Flask with convenience methods for 404:\n\n```python\nfrom flask import abort\nfrom pydantic import BaseModel\nfrom service_repository.services import BaseService as Base\n\n\nclass BaseService(Base):\n    """Class representing the base service."""\n    def __init__(self, db) -> None:\n        self.db = db\n\n    async def get_or_404(self, **kwargs):\n        """Get one instance by filter or 404 if not exists."""\n        instance = await self.get(**kwargs)\n        if not instance:\n            raise abort(404)\n        return instance\n\n    async def update_or_404(self, schema_in: BaseModel, **kwargs):\n        """Update instance or 404 if not exists."""\n        instance = await self.get_or_404(**kwargs)\n        instance = await self.update(instance=instance, schema_in=schema_in)\n        return instance\n\n    async def delete_or_404(self, **kwargs):\n        """Delete one instance by filter or 404 if not exists."""\n        await self.get_or_404(**kwargs)\n        await self.delete(**kwargs)\n```\n\n## Security\n\nIf you discover any security related issues, please email fndmiranda@gmail.com instead of using the issue tracker.\n\n## License\n\nThe MIT License (MIT). Please see [License File](LICENSE.md) for more information.',
    'author': 'Fernando Miranda',
    'author_email': 'fndmiranda@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/fndmiranda/service-repository',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



import logging

from pydantic import BaseModel

from service_repository.interfaces.service import ServiceInterface

logger = logging.getLogger(__name__)


class BaseService(ServiceInterface):
    """Class representing the abstract service."""

    _repository = None

    def __init__(self, db) -> None:
        self.db = db

    async def create(self, schema_in: BaseModel):
        """
        Create new entity and returns the saved entity instance.
        """
        create_data = schema_in.dict()
        logger.info(
            "Starting create model with={}".format(
                {
                    "service": type(self).__name__,
                    "repository": self.repository.__name__,
                    "create_data": create_data,
                }
            )
        )
        try:
            instance = await self.repository(db=self.db).create(
                schema_in=create_data
            )
            logger.info(
                "Model created successfully with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "create_data": create_data,
                        "schema_out": instance.dict(),
                    }
                )
            )

            return instance
        except Exception as exc:
            logger.error(
                "Error on create model with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "error": str(exc),
                        "create_data": create_data,
                    },
                )
            )
            raise exc

    async def update(self, instance: BaseModel, schema_in: BaseModel):
        """Update a instance."""
        update_data = schema_in.dict(exclude_unset=True)
        logger.info(
            "Starting update model with={}".format(
                {
                    "service": type(self).__name__,
                    "repository": self.repository.__name__,
                    "id": getattr(instance, "id"),
                    "update_data": update_data,
                }
            )
        )
        try:
            instance = await self.repository(db=self.db).update(
                instance=instance, schema_in=update_data
            )
            logger.info(
                "Model updated successfully with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "id": instance.id,
                        "schema_out": instance.dict(),
                    }
                )
            )
            return instance
        except Exception as exc:
            logger.error(
                "Error on update model with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "id": getattr(instance, "id"),
                        "error": str(exc),
                        "update_data": update_data,
                    }
                )
            )
            raise exc

    async def get(self, **kwargs):
        """Get one instance by filter."""
        logger.info(
            "Starting get one model with={}".format(
                {
                    "service": type(self).__name__,
                    "repository": self.repository.__name__,
                    "kwargs": kwargs,
                }
            )
        )
        try:
            instance = await self.repository(db=self.db).get(**kwargs)
            if instance:
                logger.info(
                    "Model got successfully with={}".format(
                        {
                            "service": type(self).__name__,
                            "repository": self.repository.__name__,
                            "kwargs": kwargs,
                            "schema_out": instance.dict(),
                        }
                    )
                )
                return instance
        except Exception as exc:
            logger.error(
                "Error on get one model with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "error": str(exc),
                        "kwargs": kwargs,
                    }
                )
            )
            raise exc

    async def delete(self, **kwargs):
        """Delete one instance by filter."""
        logger.info(
            "Starting delete one model with={}".format(
                {
                    "service": type(self).__name__,
                    "repository": self.repository.__name__,
                    "kwargs": kwargs,
                }
            )
        )
        try:
            await self.repository(db=self.db).delete(**kwargs)
            logger.info(
                "Model deleted successfully with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "kwargs": kwargs,
                    }
                )
            )
        except Exception as exc:
            logger.error(
                "Error on delete one model with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "error": str(exc),
                        "kwargs": kwargs,
                    }
                )
            )
            raise exc

    async def count(self, **kwargs):
        """Count instances by filter."""
        logger.info(
            "Starting count model with={}".format(
                {
                    "service": type(self).__name__,
                    "repository": self.repository.__name__,
                    "kwargs": kwargs,
                }
            )
        )
        try:
            total = await self.repository(db=self.db).count(**kwargs)
            logger.info(
                "Models counted successfully with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "kwargs": kwargs,
                        "total": total,
                    }
                )
            )
            return total
        except Exception as exc:
            logger.error(
                "Error on count model with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "error": str(exc),
                        "kwargs": kwargs,
                    }
                )
            )
            raise exc

    async def paginate(
        self,
        page: int = 1,
        per_page: int = 15,
        criteria: dict = {},
        sort: list = None,
    ):
        """Get collection of instances paginated by filter."""
        logger.info(
            "Starting paginate models with={}".format(
                {
                    "service": type(self).__name__,
                    "repository": self.repository.__name__,
                    "criteria": criteria,
                    "sort": sort,
                    "page": page,
                    "per_page": per_page,
                }
            )
        )
        try:
            pagination = await self.repository(db=self.db).paginate(
                page=page,
                per_page=per_page,
                criteria=criteria,
                sort=sort,
            )
            logger.info(
                "Models paginate successfully with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "criteria": criteria,
                        "sort": sort,
                        "items": len(pagination["items"]),
                        "per_page": pagination["per_page"],
                        "num_pages": pagination["num_pages"],
                        "page": pagination["page"],
                        "total": pagination["total"],
                    }
                )
            )
            return pagination
        except Exception as exc:
            logger.error(
                "Error on paginate models with={}".format(
                    {
                        "error": str(exc),
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "criteria": criteria,
                        "sort": sort,
                        "page": page,
                        "per_page": per_page,
                    }
                )
            )
            raise exc

    async def all(self, **kwargs):
        """Get all instances by filter."""
        logger.info(
            "Starting get models by filter with={}".format(
                {
                    "service": type(self).__name__,
                    "repository": self.repository.__name__,
                    "kwargs": kwargs,
                }
            )
        )
        try:
            instances = await self.repository(db=self.db).all(**kwargs)
            logger.info(
                "Models got successfully by filter with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "kwargs": kwargs,
                        "instances": len(instances),
                    }
                )
            )
            return instances
        except Exception as exc:
            logger.error(
                "Error on get models by filter with={}".format(
                    {
                        "service": type(self).__name__,
                        "repository": self.repository.__name__,
                        "error": str(exc),
                        "kwargs": kwargs,
                    }
                )
            )
            raise exc

    @property
    def repository(self):
        if self._repository is None:
            raise ValueError("Repository is None, set the repository")
        return self._repository

    @repository.setter
    def repository(self, value):
        self._repository = value
